from geoformat.manipulation.geolayer_manipulation import feature_list_to_geolayer
from geoformat.explore_data.print_data import print_features_data_table

time_unit_factor = {
    'h': 1 / 3600,
    "min": 1 / 60,
    "s": 1,
    "ms": 1000
}


def _format_time(time_value, time_unit=None, round_value=None):
    if time_unit is None:
        time_factor = 1
    else:
        time_factor = time_unit_factor[time_unit]

    format_time = time_value * time_factor

    if round_value is not None:
        format_time = round(format_time, round_value)

    return format_time


class Timer:

    def __init__(self, timer_name):
        self.name = timer_name
        self.time_ms = 0
        self.count_call = 0
        self.function_in_it = {}
        self.decorator_time = 0

    def add_decorator_time(self, execute_time):
        self.decorator_time += execute_time

    def add_child_timer(self, child_timer_object):
        self.function_in_it[child_timer_object.name] = child_timer_object

    def add_time(self, execute_time):
        self.time_ms += execute_time
        self.count_call += 1

    def minus_time(self, execute_time):
        self.time_ms -= execute_time

    def get_children_name(self):
        return self.function_in_it.keys()

    def get_child_timer(self, child_name):
        if child_name in self.function_in_it:
            return self.function_in_it[child_name]
        else:
            raise Exception("This is not a child of {}".format(self.name))

    def _print_function_timer(self, indent=0, time_unit='ms'):
        if self.function_in_it:
            for function_timer_name, function_timer in self.function_in_it.items():
                function_timer._print_function_timer(indent=indent + 1, time_unit=time_unit)

    def _print_timer(self, indent, time_unit='ms', round_value=3):
        if indent:
            indent_txt = '\t' * indent
        else:
            indent_txt = ''

        str_return = '{indent_txt}{function_name}:'.format(
            indent_txt=indent_txt,
            function_name=self.name
        )
        if self.count_call:
            format_time = round(self.time_ms * time_unit_factor[time_unit], round_value)
            str_return = '{str_return}time ({time_unit}): {time}, nb call: {nb_call}, time by call, {time_call}'.format(
                str_return=str_return,
                time=format_time,
                time_unit=time_unit,
                nb_call=self.count_call,
                time_call=round(format_time / self.count_call, round_value)
            )

        # add decorator time
        decorator_time = round(self.decorator_time * time_unit_factor[time_unit], round_value)
        str_return += ', decorator time: {decorator_time}'.format(decorator_time=decorator_time)
        return str_return

    def _get_feature(self, time_field_list, indent=0, path=None, path_tree_dict=None):

        if path is None:
            path = []

        if path_tree_dict is None:
            path_tree_dict = {}

        if indent:
            indent_txt = '\t' * indent
        else:
            indent_txt = ''

        path = tuple(path + [self.name])
        feature = {
            "attributes": {
                'function_name': '{indent_txt}{name}'.format(
                    indent_txt=indent_txt.expandtabs(),
                    name=self.name
                ),
                'path': path
            }
        }
        if self.count_call:
            time_value = self.time_ms
            format_time_field_name, format_decorator_time_field_name, format_mean_time_field_name, format_true_time_field_name, format_true_mean_time_field_name = time_field_list
            feature['attributes'][format_time_field_name] = time_value
            feature['attributes'][format_decorator_time_field_name] = self.decorator_time
            feature['attributes']['count_call'] = self.count_call
            feature['attributes'][format_mean_time_field_name] = time_value / self.count_call

            # compute true time
            if path in path_tree_dict:
                feature['attributes'][format_true_time_field_name] = time_value - path_tree_dict[path]
                feature['attributes'][format_true_mean_time_field_name] = feature['attributes'][format_true_time_field_name] / self.count_call
        return feature

    def _get_features(self, time_field_list, indent=0, path=[], path_tree_dict={}):
        if self.function_in_it:
            path = path + [self.name]
            for function_timer_name, function_timer in self.function_in_it.items():
                yield function_timer._get_feature(time_field_list=time_field_list, indent=indent, path=path,
                                                  path_tree_dict=path_tree_dict)
                for feature in function_timer._get_features(time_field_list=time_field_list, indent=indent + 1, path=path,
                                                            path_tree_dict=path_tree_dict):
                    yield feature

    def _get_associate_decorator_time_for_function_path(self, function_dict=None, function_path=None):
        """
        Create dict with path in key and decorator time associate to given path
        :param function_dict:
        :param function_path:
        :return:
        """

        if function_dict is None:
            function_dict = {}
        if function_path is None:
            function_path = []

        # function_path.append(self.name)
        function_dict[tuple(function_path + [self.name])] = {'decorator_time': self.decorator_time}

        if self.function_in_it:
            function_path = function_path + [self.name]
            for child_name, child in self.function_in_it.items():
                child_function_dict = child._get_associate_decorator_time_for_function_path(
                    function_dict=function_dict,
                    function_path=function_path)
                for key, value in child_function_dict.items():
                    function_dict[key] = value

        return function_dict

    def _get_sum_of_decorator_time_for_function_path(self, function_path_decorator_time):
        """
        Add decorator time from all child to function_path

        :param function_path_decorator_time:
        :return:
        """
        path_list = function_path_decorator_time.keys()
        path_tree_dict = {}
        for path_tuple in path_list:
            request_path_list = []
            for path_value in path_tuple:
                request_path_list.append(path_value)
                request_path_tuple = tuple(request_path_list)
                len_request_path_tuple = len(request_path_list)
                if tuple(request_path_tuple) not in path_tree_dict:
                    # create new entry
                    path_tree_dict[request_path_tuple] = 0  # zero because value is added in second loop
                    # loop on path_list to add child decorator time
                    for path_tuple_2 in path_list:
                        # if child path
                        if request_path_tuple == path_tuple_2[:len_request_path_tuple] and len(
                                path_tuple_2) > len_request_path_tuple:
                            # add child decorator time
                            path_tree_dict[request_path_tuple] += function_path_decorator_time[path_tuple_2][
                                'decorator_time']

        return path_tree_dict

    def _get_format_field_time(self, features_list, time_field_list, time_unit, round_value=3):
        """
        format time field to given time unit

        :param features_list:
        :param time_field_list:
        :param time_unit:
        :return:
        """

        for feature in features_list:
            feature_attributes = feature['attributes']
            del feature_attributes['path']
            for time_field in time_field_list:
                if time_field in feature_attributes:
                    feature_attributes[time_field] = _format_time(
                        feature_attributes[time_field],
                        time_unit=time_unit,
                        round_value=round_value
                    )

        return features_list

    def print(self, time_unit='ms', round_value=3):

        # we create a dictionary that store decorator time for each call
        function_path_decorator_time = self._get_associate_decorator_time_for_function_path()
        # we create a dictionary that sum decorator time for function
        path_tree_dict = self._get_sum_of_decorator_time_for_function_path(
            function_path_decorator_time=function_path_decorator_time
        )
        # create features that store time information
        # compute time field
        format_time_field_name = 'time_{}'.format(time_unit)
        format_mean_time_field_name = 'mean_time_by_call_{}'.format(time_unit)
        format_decorator_time_field_name = 'decorator_time_{}'.format(time_unit)
        format_true_time_field_name = 'true_{}'.format(format_time_field_name)
        format_true_mean_time_field_name = 'true_{}'.format(format_mean_time_field_name)
        time_field_list = [
            format_time_field_name,
            format_mean_time_field_name,
            format_decorator_time_field_name,
            format_true_time_field_name,
            format_true_mean_time_field_name
        ]
        indent = 0
        features_list = [self._get_feature(time_field_list=time_field_list, indent=indent, path=[], path_tree_dict=path_tree_dict)]
        indent = 1
        features_list += [feature for feature in
                          self._get_features(time_field_list=time_field_list, indent=indent, path=[], path_tree_dict=path_tree_dict)]

        # format time field in given unit
        features_list = self._get_format_field_time(features_list, time_field_list, time_unit, round_value=round_value)

        # we print result
        print(print_features_data_table(geolayer=feature_list_to_geolayer(features_list, geolayer_name='timer')))



from geoformat.conf.error_messages import (
    geolayer_attributes_missing,
    field_missing,
    geometry_ref_geolayer,
)
from geoformat.index.attributes.hash import create_attribute_index
from geoformat.processing.data.join.merge_objects import merge_metadata, merge_feature
from geoformat.processing.data.union import union_geolayer
from geoformat.manipulation.geolayer_manipulation import (
    add_attributes_index,
    check_attributes_index,
    delete_attributes_index,
)

from geoformat.processing.data.clauses import clause_where
from geoformat.conversion.geolayer_conversion import create_geolayer_from_i_feat_list


def _check_input_geolayer_data(geolayer_a, geolayer_b, on_field_a, on_field_b):
    # check input data
    if "fields" in geolayer_a["metadata"]:
        if on_field_a not in geolayer_a["metadata"]["fields"]:
            raise Exception(field_missing.format(field_name=on_field_a))
    else:
        raise Exception(
            geolayer_attributes_missing.format(
                geolayer_name=geolayer_a["metadata"]["name"]
            )
        )

    if "fields" in geolayer_b["metadata"]:
        if on_field_b not in geolayer_b["metadata"]["fields"]:
            raise Exception(field_missing.format(field_name=on_field_b))
    else:
        raise Exception(
            geolayer_attributes_missing.format(
                geolayer_name=geolayer_b["metadata"]["name"]
            )
        )


def _name_output_geolayer(output_geolayer_name, geolayer_a_name, geolayer_b_name):

    return output_geolayer_name or "{geolayer_a_name}_join_{geolayer_b_name}".format(
        geolayer_a_name=geolayer_a_name,
        geolayer_b_name=geolayer_b_name,
    )


def _join_create_output_geolayer(
    geolayer_a,
    geolayer_b,
    output_geolayer_name,
    field_name_filter_a,
    field_name_filter_b,
    geometry_ref,
    rename_output_field_from_geolayer_a,
    rename_output_field_from_geolayer_b,
):

    if geometry_ref == "geolayer_a":
        metadata_geometry_ref = "metadata_a"
    elif geometry_ref == "geolayer_b":
        metadata_geometry_ref = "metadata_b"
    else:
        raise Exception(geometry_ref_geolayer)

    # merge geolayer metadata
    merge_metadata_result = merge_metadata(
        metadata_a=geolayer_a["metadata"],
        metadata_b=geolayer_b["metadata"],
        geolayer_name=_name_output_geolayer(
            output_geolayer_name=output_geolayer_name,
            geolayer_a_name=geolayer_a["metadata"]["name"],
            geolayer_b_name=geolayer_b["metadata"]["name"],
        ),
        field_name_filter_a=field_name_filter_a,
        field_name_filter_b=field_name_filter_b,
        geometry_ref=metadata_geometry_ref,
        rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
        rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
    )
    output_metadata = merge_metadata_result["metadata"]
    rename_field_correspondence_a = merge_metadata_result["fields_correspondance_a"]
    rename_field_correspondence_b = merge_metadata_result["fields_correspondance_b"]

    # creating new geolayer
    output_geolayer = {"metadata": output_metadata, "features": {}}

    return (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
    )


def _get_attributes_index(geolayer, field_name_to_index):

    # check if geolayer_join have an attributes index on on_field_join (if not we create it)
    field_index = (
        geolayer["metadata"]
        .get("index", {})
        .get("attributes", {})
        .get(field_name_to_index, None)
    )
    if field_index is None:
        field_index = create_attribute_index(
            geolayer=geolayer, field_name=field_name_to_index
        )

    return field_index


def _prepare_join(
    geolayer_a,
    geolayer_b,
    output_geolayer_name,
    on_field_a,
    on_field_b,
    field_name_filter_a,
    field_name_filter_b,
    rename_output_field_from_geolayer_a,
    rename_output_field_from_geolayer_b,
    geometry_ref,
):
    # check input geolayer
    _check_input_geolayer_data(
        geolayer_a=geolayer_a,
        geolayer_b=geolayer_b,
        on_field_a=on_field_a,
        on_field_b=on_field_b,
    )

    # create output geolayer, output_metadata and rename field correspondence
    (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
    ) = _join_create_output_geolayer(
        geolayer_a=geolayer_a,
        geolayer_b=geolayer_b,
        output_geolayer_name=output_geolayer_name,
        field_name_filter_a=field_name_filter_a,
        field_name_filter_b=field_name_filter_b,
        geometry_ref=geometry_ref,
        rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
        rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
    )

    feature_geometry_ref = "feature_{}"
    if geometry_ref == "geolayer_a":
        feature_geometry_ref = feature_geometry_ref.format("a")
    elif geometry_ref == "geolayer_b":
        feature_geometry_ref = feature_geometry_ref.format("b")
    else:
        raise Exception(geometry_ref_geolayer)

    return (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
        feature_geometry_ref,
    )


def _left_right_join(
    geolayer_left_or_right,
    geolayer_join,
    on_field_left_or_right,
    on_field_join,
    output_metadata,
    field_name_filter_left_or_right,
    field_name_filter_join,
    rename_field_correspondence_left_or_right,
    rename_field_correspondence_join,
    geometry_ref,
    output_geolayer,
):
    # check if geolayer_join have an attributes index on on_field_join (if not we create it)
    on_field_join_index = _get_attributes_index(
        geolayer=geolayer_join, field_name_to_index=on_field_join
    )

    # loop on each feature and make join
    output_geolayer_i_feat = 0
    for i_feat, feature_left_or_right in geolayer_left_or_right["features"].items():
        feature_join_list = []
        if feature_left_right_attributes := feature_left_or_right.get("attributes"):
            # get field value
            on_field_left_right_value = feature_left_right_attributes.get(
                on_field_left_or_right
            )
            if on_field_left_right_value is not None:
                # get join field value
                if on_field_join_i_feat_list := on_field_join_index["index"].get(
                    on_field_left_right_value
                ):
                    for geolayer_join_i_feat in on_field_join_i_feat_list:
                        feature_join_list.append(
                            geolayer_join["features"][geolayer_join_i_feat]
                        )
        # if no feature to join add empty dict
        if not feature_join_list:
            feature_join_list.append({})

        # merge feature_left_or_right with feature_join and add to output_geolayer
        for feature_join in feature_join_list:
            new_feature = merge_feature(
                feature_a=feature_left_or_right,
                feature_b=feature_join,
                merge_metadata=output_metadata,
                field_name_filter_a=field_name_filter_left_or_right,
                field_name_filter_b=field_name_filter_join,
                rename_fields_a=rename_field_correspondence_left_or_right,
                rename_fields_b=rename_field_correspondence_join,
                geometry_ref=geometry_ref,
            )

            output_geolayer["features"][output_geolayer_i_feat] = new_feature
            output_geolayer_i_feat += 1

    return output_geolayer


def join(
    geolayer_a,
    geolayer_b,
    on_field_a,
    on_field_b,
    output_geolayer_name=None,
    field_name_filter_a=None,
    field_name_filter_b=None,
    rename_output_field_from_geolayer_a="auto",
    rename_output_field_from_geolayer_b="auto",
    geometry_ref="geolayer_a",
):
    """
     Make an attributes join between two geolayer.
         - keep only feature with matching on_field_a and on_field_b

       *  *   *  *
     *     *~~*     *
    *  A  *~~~~*  B  *
    *     *~~~~*     *
     *     *~~*     *
        *  *  *  *

     :param geolayer_a: first geolayer to join.
     :param geolayer_b: second geolayer to join.
     :param on_field_a: field of geolayer_a on which we make the join.
     :param on_field_b: field of geolayer_b on which we make the join.
     :param output_geolayer_name: name of new geolayer
     :param field_name_filter_a: field from geolayer_a that we keep.
     :param field_name_filter_b: field from geolayer_b that we keep.
     :param rename_output_field_from_geolayer_a: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param rename_output_field_from_geolayer_b: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param geometry_ref: if there is geometry in geolayer_a and/or geolayer_b from which geolayer geometry is keep ? (
    'geolayer_a' or 'geolayer_b').
     :return: geolayer_a join by geolayer_b
    """
    (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
        feature_geometry_ref,
    ) = _prepare_join(
        geolayer_a,
        geolayer_b,
        output_geolayer_name,
        on_field_a,
        on_field_b,
        field_name_filter_a,
        field_name_filter_b,
        rename_output_field_from_geolayer_a,
        rename_output_field_from_geolayer_b,
        geometry_ref,
    )

    # create index
    on_field_a_index = _get_attributes_index(
        geolayer=geolayer_a, field_name_to_index=on_field_a
    )
    on_field_b_index = _get_attributes_index(
        geolayer=geolayer_b, field_name_to_index=on_field_b
    )

    # get common key value
    common_value_set = set(on_field_a_index["index"].keys()).intersection(
        set(on_field_b_index["index"].keys())
    )
    # delete none value
    common_value_set = common_value_set - {None}
    # get i_feat
    i_feat_list = []
    for common_value in common_value_set:
        for i_feat in on_field_a_index["index"][common_value]:
            i_feat_list.append(i_feat)

    output_geolayer_i_feat = 0
    for i_feat in sorted(i_feat_list):
        feature_a = geolayer_a["features"][i_feat]
        feature_a_field_value = feature_a.get("attributes").get(on_field_a)
        for i_feat_b in sorted(on_field_b_index["index"][feature_a_field_value]):
            feature_b = geolayer_b["features"][i_feat_b]
            new_feature = merge_feature(
                feature_a,
                feature_b,
                merge_metadata=output_metadata,
                field_name_filter_a=field_name_filter_a,
                field_name_filter_b=field_name_filter_b,
                rename_fields_a=rename_field_correspondence_a,
                rename_fields_b=rename_field_correspondence_b,
                geometry_ref=feature_geometry_ref,
            )
            output_geolayer["features"][output_geolayer_i_feat] = new_feature
            output_geolayer_i_feat += 1

    return output_geolayer


def join_full(
    geolayer_a,
    geolayer_b,
    on_field_a,
    on_field_b,
    output_geolayer_name=None,
    field_name_filter_a=None,
    field_name_filter_b=None,
    rename_output_field_from_geolayer_a="auto",
    rename_output_field_from_geolayer_b="auto",
    geometry_ref="geolayer_a",
):
    """
     Make an attributes full join between two geolayer.
     - keep all feature with matching or not on_field_a and on_field_b.
     - If no matching the field from geolayer_a or geolayer_b that not match have None value.

           *~~*   *~~*
         *~~~~~*~~*~~~~~*
        *~~A~~*~~~~*~~B~~*
        *~~~~~*~~~~*~~~~~*
         *~~~~~*~~*~~~~~*
            *~~*  *~~*

     :param geolayer_a: first geolayer to join.
     :param geolayer_b: second geolayer to join.
     :param on_field_a: field of geolayer_a on which we make the join.
     :param on_field_b: field of geolayer_b on which we make the join.
     :param output_geolayer_name: name of new geolayer
     :param field_name_filter_a: field from geolayer_a that we keep.
     :param field_name_filter_b: field from geolayer_b that we keep.
     :param rename_output_field_from_geolayer_a: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param rename_output_field_from_geolayer_b: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param geometry_ref: if there is geometry in geolayer_a and/or geolayer_b from which geolayer geometry is keep ? (
    'geolayer_a' or 'geolayer_b').
     :return: geolayer_a join by geolayer_b
    """

    # create or get attributes index for on_field_a and on_field_b
    delete_index_a = False
    if check_attributes_index(geolayer=geolayer_a, field_name=on_field_a) is False:
        geolayer_a_on_field_index = create_attribute_index(
            geolayer=geolayer_a, field_name=on_field_a
        )
        add_attributes_index(
            geolayer=geolayer_a, field_name=on_field_a, index=geolayer_a_on_field_index
        )
        delete_index_a = True
    else:
        geolayer_a_on_field_index = geolayer_a["metadata"]["index"]["attributes"][
            on_field_a
        ]

    delete_index_b = False
    if check_attributes_index(geolayer=geolayer_b, field_name=on_field_b) is False:
        geolayer_b_on_field_index = create_attribute_index(
            geolayer=geolayer_b, field_name=on_field_b
        )
        add_attributes_index(
            geolayer=geolayer_b, field_name=on_field_b, index=geolayer_b_on_field_index
        )
        delete_index_b = True
    else:
        geolayer_b_on_field_index = geolayer_a["metadata"]["index"]["attributes"][
            on_field_b
        ]

    geolayer_a_on_field_value_set = set(geolayer_a_on_field_index["index"].keys())
    geolayer_b_on_field_value_set = set(geolayer_b_on_field_index["index"].keys())

    # check i_feat for left join only
    value_in_a_not_in_b = geolayer_a_on_field_value_set - geolayer_b_on_field_value_set

    # check i_feat for join only
    value_in_a_and_b = geolayer_a_on_field_value_set.intersection(
        geolayer_b_on_field_value_set
    )

    value_in_b_not_in_a = geolayer_b_on_field_value_set - geolayer_a_on_field_value_set

    # make join between geolayers
    union_geolayer_list = []
    if value_in_a_not_in_b:
        left_join_geolayer = join_left(
            geolayer_a=geolayer_a,
            geolayer_b=geolayer_b,
            on_field_a=on_field_a,
            on_field_b=on_field_b,
            output_geolayer_name=output_geolayer_name,
            field_name_filter_a=field_name_filter_a,
            field_name_filter_b=field_name_filter_b,
            rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
            rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
            geometry_ref=geometry_ref,
        )
        union_geolayer_list.append(left_join_geolayer)

    if (
        value_in_a_and_b
    ):  # and (value_in_a_not_in_b == set() and value_in_b_not_in_a == set()):
        if value_in_a_not_in_b == set() and value_in_b_not_in_a == set():
            join_geolayer = join(
                geolayer_a=geolayer_a,
                geolayer_b=geolayer_b,
                on_field_a=on_field_a,
                on_field_b=on_field_b,
                output_geolayer_name=output_geolayer_name,
                field_name_filter_a=field_name_filter_a,
                field_name_filter_b=field_name_filter_b,
                rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
                rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
                geometry_ref=geometry_ref,
            )
            union_geolayer_list.append(join_geolayer)

            if None in value_in_a_and_b:
                geolayer_a_whith_only_none_i_feat_list = clause_where(
                    geolayer=geolayer_a,
                    field_name=on_field_a,
                    predicate="=",
                    values=None,
                )
                geolayer_a_with_only_none = create_geolayer_from_i_feat_list(
                    geolayer=geolayer_a,
                    i_feat_list=geolayer_a_whith_only_none_i_feat_list,
                )
                geolayer_b_whith_only_none_i_feat_list = clause_where(
                    geolayer=geolayer_b,
                    field_name=on_field_b,
                    predicate="=",
                    values=None,
                )
                geolayer_b_whith_only_none = create_geolayer_from_i_feat_list(
                    geolayer=geolayer_b,
                    i_feat_list=geolayer_b_whith_only_none_i_feat_list,
                )

                geolayer_none_left = join_left(
                    geolayer_a=geolayer_a_with_only_none,
                    geolayer_b=geolayer_b_whith_only_none,
                    on_field_a=on_field_a,
                    on_field_b=on_field_b,
                    output_geolayer_name=output_geolayer_name,
                    field_name_filter_a=field_name_filter_a,
                    field_name_filter_b=field_name_filter_b,
                    rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
                    rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
                    geometry_ref=geometry_ref,
                )
                union_geolayer_list.append(geolayer_none_left)

                geolayer_none_right = join_right(
                    geolayer_a=geolayer_a_with_only_none,
                    geolayer_b=geolayer_b_whith_only_none,
                    on_field_a=on_field_a,
                    on_field_b=on_field_b,
                    output_geolayer_name=output_geolayer_name,
                    field_name_filter_a=field_name_filter_a,
                    field_name_filter_b=field_name_filter_b,
                    rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
                    rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
                    geometry_ref=geometry_ref,
                )
                union_geolayer_list.append(geolayer_none_right)

    if value_in_b_not_in_a:
        right_join_geolayer = join_right(
            geolayer_a=geolayer_a,
            geolayer_b=geolayer_b,
            on_field_a=on_field_a,
            on_field_b=on_field_b,
            output_geolayer_name=output_geolayer_name,
            field_name_filter_a=field_name_filter_a,
            field_name_filter_b=field_name_filter_b,
            rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
            rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
            geometry_ref=geometry_ref,
        )
        union_geolayer_list.append(right_join_geolayer)

    # delete duplicate values

    # union join
    if len(union_geolayer_list) > 1:
        output_geolayer = union_geolayer(
            union_geolayer_list, geolayer_name=output_geolayer_name, serialize=False
        )
    else:
        output_geolayer = union_geolayer_list[0]

    # name output geolayer
    output_geolayer["metadata"]["name"] = _name_output_geolayer(
        output_geolayer_name=output_geolayer_name,
        geolayer_a_name=geolayer_a["metadata"]["name"],
        geolayer_b_name=geolayer_b["metadata"]["name"],
    )

    # delete created index if not exists originally
    if delete_index_a is True:
        delete_attributes_index(geolayer=geolayer_a, field_name=on_field_a)
    if delete_index_b is True:
        delete_attributes_index(geolayer=geolayer_b, field_name=on_field_b)

    return output_geolayer


def join_left(
    geolayer_a,
    geolayer_b,
    on_field_a,
    on_field_b,
    output_geolayer_name=None,
    field_name_filter_a=None,
    field_name_filter_b=None,
    rename_output_field_from_geolayer_a="auto",
    rename_output_field_from_geolayer_b="auto",
    geometry_ref="geolayer_a",
):
    """
     Make an attributes left join between two geolayer.
         - keep all features from geolayer_a and only features from geolayer_b with matching on_field_a and on_field_b

       *~~*   *  *
     *~~~~~*~~*     *
    *~~A~~*~~~~*  B  *
    *~~~~~*~~~~*     *
     *~~~~~*~~*     *
        *~~*  *  *

     :param geolayer_a: first geolayer to join.
     :param geolayer_b: second geolayer to join.
     :param on_field_a: field of geolayer_a on which we make the join.
     :param on_field_b: field of geolayer_b on which we make the join.
     :param output_geolayer_name: name of new geolayer
     :param field_name_filter_a: field from geolayer_a that we keep.
     :param field_name_filter_b: field from geolayer_b that we keep.
     :param rename_output_field_from_geolayer_a: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param rename_output_field_from_geolayer_b: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param geometry_ref: if there is geometry in geolayer_a and/or geolayer_b from which geolayer geometry is keep ? (
    'geolayer_a' or 'geolayer_b').
     :return: geolayer_a join by geolayer_b"""

    (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
        feature_geometry_ref,
    ) = _prepare_join(
        geolayer_a=geolayer_a,
        geolayer_b=geolayer_b,
        output_geolayer_name=output_geolayer_name,
        on_field_a=on_field_a,
        on_field_b=on_field_b,
        field_name_filter_a=field_name_filter_a,
        field_name_filter_b=field_name_filter_b,
        rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
        rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
        geometry_ref=geometry_ref,
    )

    output_geolayer = _left_right_join(
        geolayer_left_or_right=geolayer_a,
        geolayer_join=geolayer_b,
        on_field_left_or_right=on_field_a,
        on_field_join=on_field_b,
        output_metadata=output_metadata,
        field_name_filter_left_or_right=field_name_filter_a,
        field_name_filter_join=field_name_filter_b,
        rename_field_correspondence_left_or_right=rename_field_correspondence_a,
        rename_field_correspondence_join=rename_field_correspondence_b,
        geometry_ref=feature_geometry_ref,
        output_geolayer=output_geolayer,
    )

    return output_geolayer


def join_right(
    geolayer_a,
    geolayer_b,
    on_field_a,
    on_field_b,
    output_geolayer_name=None,
    field_name_filter_a=None,
    field_name_filter_b=None,
    rename_output_field_from_geolayer_a="auto",
    rename_output_field_from_geolayer_b="auto",
    geometry_ref="geolayer_b",
):
    """
     Make an attributes right join between two geolayer.
         - keep all features from geolayer_b and only features from geolayer_a with matching on_field_a and on_field_b

       *  *   *~~*
     *     *~~*~~~~~*
    *  A  *~~~~*~~B~~*
    *     *~~~~*~~~~~*
     *     *~~*~~~~~*
        *  *  *~~*

     :param geolayer_a: first geolayer to join.
     :param geolayer_b: second geolayer to join.
     :param on_field_a: field of geolayer_a on which we make the join.
     :param on_field_b: field of geolayer_b on which we make the join.
     :param output_geolayer_name: name of new geolayer
     :param field_name_filter_a: field from geolayer_a that we keep.
     :param field_name_filter_b: field from geolayer_b that we keep.
     :param rename_output_field_from_geolayer_a: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param rename_output_field_from_geolayer_b: field from geolayer that we want to rename in output geolayer
     (to avoid field with same name between geolayer_a and geolayer_b). By default, the renaming is automatic, but
     you can give a dict with thi pattern :
         {'old_field_name_x': 'new_field_name_x' 'old_field_name_y': 'new_field_name_y'}
     :param geometry_ref: if there is geometry in geolayer_a and/or geolayer_b from which geolayer geometry is keep ? (
    'geolayer_a' or 'geolayer_b').
     :return: geolayer_a join by geolayer_b
    """
    (
        output_geolayer,
        output_metadata,
        rename_field_correspondence_a,
        rename_field_correspondence_b,
        feature_geometry_ref,
    ) = _prepare_join(
        geolayer_a=geolayer_a,
        geolayer_b=geolayer_b,
        output_geolayer_name=output_geolayer_name,
        on_field_a=on_field_a,
        on_field_b=on_field_b,
        field_name_filter_a=field_name_filter_a,
        field_name_filter_b=field_name_filter_b,
        rename_output_field_from_geolayer_a=rename_output_field_from_geolayer_a,
        rename_output_field_from_geolayer_b=rename_output_field_from_geolayer_b,
        geometry_ref=geometry_ref,
    )
    if feature_geometry_ref == "feature_b":
        feature_geometry_ref = "feature_a"
    else:
        feature_geometry_ref = "feature_b"

    output_geolayer = _left_right_join(
        geolayer_left_or_right=geolayer_b,
        geolayer_join=geolayer_a,
        on_field_left_or_right=on_field_b,
        on_field_join=on_field_a,
        output_metadata=output_metadata,
        field_name_filter_left_or_right=field_name_filter_b,
        field_name_filter_join=field_name_filter_a,
        rename_field_correspondence_left_or_right=rename_field_correspondence_b,
        rename_field_correspondence_join=rename_field_correspondence_a,
        geometry_ref=feature_geometry_ref,
        output_geolayer=output_geolayer,
    )

    return output_geolayer



import copy
import datetime
import sys

from geoformat.conf.format_data import value_to_iterable_value, is_hexadecimal

from geoformat.conf.fields_variable import none_value_pattern

from geoformat.conversion.geometry_conversion import (
    geometry_to_wkb,
    wkb_to_geometry,
    geometry_to_bbox,
    geometry_to_geometry_collection
)
from geoformat.geoprocessing.connectors.predicates import bbox_intersects_bbox
from geoformat.geoprocessing.geoparameters.bbox import bbox_union


def feature_serialize(feature):
    """
    Serialize feature.
        "attributes" are converted to string. TODO serialize in bytes.
        "geometry' are converted to WKB.

    :param feature: feature that we want to convert
    :return: feature serialized
    """
    serialized_feature = {}
    if 'attributes' in feature:
        serialized_feature["attributes"] = str(feature['attributes'])

    if 'geometry' in feature:
        serialized_feature["geometry"] = geometry_to_wkb(feature['geometry'])

    return serialized_feature


def feature_deserialize(serialized_feature, bbox=True):
    """
    Convert serialized feature to non serialized feature

    :param serialized_feature: feature serialized
    :param bbox: True if you want to add bbox information geometry key
    :return: non serialized geometry
    """
    feature = {}
    if "attributes" in serialized_feature:
        attributes = eval(serialized_feature["attributes"])
        feature["attributes"] = attributes
    if "geometry" in serialized_feature:
        geometry = wkb_to_geometry(serialized_feature["geometry"], bbox=bbox)
        feature["geometry"] = geometry

    return feature


def features_geometry_ref_scan(
        geolayer_or_feature_list,
        geometry_type_filter=None,
        bbox_filter=None,
        extent=True
):
    """
    Loop on each features on a geolayer or a list and deduce geometries metadata from it.

    :param geolayer_or_feature_list: geolayer or features list to scan
    :param geometry_type_filter: filter on geolayer_or_feature_list only geometry type sepcified on this variable
           (can be a list)
    :param bbox_filter: filter on features geometry that intersects bbox (can be a list)
    :param extent: if True add geolayer extent to geometry metadata
    :return: geometry scan result
    """
    # init
    geometry_type_filter = value_to_iterable_value(geometry_type_filter, set)
    if isinstance(bbox_filter, (list, tuple)):
        if isinstance(bbox_filter[0], (int, float)):
            bbox_filter = [bbox_filter]
    bbox_filter = value_to_iterable_value(bbox_filter, set)
    geometry_type_set = set([])

    # check if input data is geolayer or feature
    if isinstance(geolayer_or_feature_list, list):
        is_geolayer = False
        feature_list = geolayer_or_feature_list
    elif isinstance(geolayer_or_feature_list, dict):
        is_geolayer = True
        feature_list = geolayer_or_feature_list["features"]
    else:
        raise Exception('geolayer_or_feature_list must be a list of features or')

    extent_value = None
    for i, feature in enumerate(feature_list):
        if is_geolayer:
            feature = geolayer_or_feature_list["features"][feature]

        # check if geometry
        if "geometry" in feature:
            geometry = feature["geometry"]
            # get geometry type
            geometry_type = geometry['type']

            # check geometry type
            if geometry_type_filter:
                # if geometry not in geometry_type_filter we loop on next feature
                if geometry_type not in geometry_type_filter:
                    continue

            # check bbox
            if bbox_filter:
                if "bbox" in geometry:
                    feature_bbox = geometry['bbox']
                else:
                    feature_bbox = geometry_to_bbox(geometry)

                geometry_in_bbox = False
                # loop on each bbox from bbox_filter
                for bbox in bbox_filter:
                    if bbox_intersects_bbox(bbox, feature_bbox):
                        geometry_in_bbox = True
                        break

                # if geometry not in bbox we loop on next feature
                if geometry_in_bbox is False:
                    continue

            # add geometry type in geometry_type_set
            geometry_type_set.update([geometry_type])

            # if extent option in True
            if extent:
                if 'bbox' in feature:
                    bbox = feature['bbox']
                else:
                    bbox = geometry_to_bbox(geometry)
                # compute extent
                if extent_value:
                    extent_value = bbox_union(bbox, extent_value)
                else:
                    extent_value = bbox

    # create geometry ref metadata dict
    if not geometry_type_set:
        geometry_type_set = None

    geometry_return = {'type': geometry_type_set, 'extent': extent_value}

    return geometry_return


def check_if_value_is_from_datetime_lib(datetime_type, value, try_to_force):
    """
    Return True if value is instance of indicated datetime_type

    :param datetime_type: datetime type that we want test
    :param value: value to test
    :param try_to_force: True if you want force input value
    :return: True / False
    """

    if isinstance(value, datetime_type):
        return_value = True
    else:
        return_value = False

    if try_to_force is True and return_value is False:
        if sys.version_info >= (3, 7):
            try:
                value_from_isoformat = datetime_type.fromisoformat(value)
                if value_from_isoformat:
                    return_value = True

            except (ValueError, TypeError):
                return_value = False
        else:
            return_value = False

    return return_value


def check_if_value_is_date(value, try_to_force):
    """
    check if value is date type

    :param value: value to test
    :param try_to_force: True if you want to force value to datetime type date
    :return:
    """
    return check_if_value_is_from_datetime_lib(datetime_type=datetime.date, value=value, try_to_force=try_to_force)


def check_if_value_is_datetime(value, try_to_force):
    """
    check if value is datetime type

    :param value: value to test
    :param try_to_force: True if you want to force value to datetime type datetime
    :return:
    """
    return check_if_value_is_from_datetime_lib(datetime_type=datetime.datetime, value=value, try_to_force=try_to_force)


def check_if_value_is_time(value, try_to_force):
    """
    check if value is datetime time

    :param value: value to test
    :param try_to_force: True if you want to force value to datetime type time
    :return:
    """
    return check_if_value_is_from_datetime_lib(datetime_type=datetime.time, value=value, try_to_force=try_to_force)


def return_if_value_is_time_date_or_datetime(value, try_to_force):
    """
    Determine if value is date / time or datetime returns None if none of the this cases.

    :param value: value to test
    :param try_to_force: True if you want to force value to datetime type
    :return: datetime type (date / time or datetime).
    """

    if isinstance(value, (datetime.date, datetime.time, datetime.datetime)):
        return_value = type(value)
    else:
        is_date = check_if_value_is_date(value=value, try_to_force=try_to_force)
        is_time = check_if_value_is_time(value=value, try_to_force=try_to_force)
        is_datetime = check_if_value_is_datetime(value=value, try_to_force=try_to_force)

        if is_date is False and is_time is False and is_datetime is True:
            # datetime is compatible with date (that why we don't test it).
            return_value = datetime.datetime
        elif is_date is True and is_time is False and is_datetime is True:
            return_value = datetime.date
        elif is_date is False and is_time is True and is_datetime is False:
            return_value = datetime.time
        elif is_date is True and is_time is False and is_datetime is False:
            return_value = datetime.date
        else:
            return_value = None

    return return_value


def features_fields_type_scan(
    geolayer_or_feature_list,
    field_name_filter=None,
    try_to_force_type=False,
    fields_index=True,
    none_value_pattern=none_value_pattern
):
    """
    Loop on each features on a geolayer or a list and deduce fields metadata from it.

    :param geolayer_or_feature_list: geolayer_or_feature_list: geolayer or features list to scan
    :param field_name_filter: filter only on specified field_name (can be a list)
    :param try_to_force_type: option that can force the Type of value on geolayer or features list
    :param fields_index: if True add field index position on field_name contians in geolayer or features list
    :return: fields scan result
    """

    def define_field_type(field_dict, field_value, field_try_to_force_type=False):
        """
        Return from a given list deduce type of data's in it.
        To work this function need field dict that is the summary of all data type for the field

        TODO : big refacto to do here :
                - work with sub-functions by field type (which takes try_to_force as parameter)

        :param field_dict: field's dictionary that type you want to know
        :param field_value: value_to_force in field
        :param field_try_to_force_type:
        :return: return edited field_dict with parameters that we deduce from field_value
        """

        def force_value(value_to_force, force_type):
            """
            This function force a given value to an other type.
            If type is incompatible function return None

            :param value_to_force: value to recast
            :param force_type:  type that we want to recast value
            :return: new value
            """

            try:
                if force_type == bool:
                    if isinstance(value_to_force, str):
                        if value_to_force.lower() == 'true':
                            forced_value = True
                        elif value_to_force.lower() == 'false':
                            forced_value = False
                        else:
                            forced_value = None
                    elif isinstance(value_to_force, int):
                        if value_to_force == 1:
                            forced_value = True
                        elif value_to_force == 0:
                            forced_value = False
                        else:
                            forced_value = None
                    elif isinstance(value_to_force, bool):
                        return value_to_force
                    else:
                        forced_value = None
                else:
                    forced_value = force_type(value_to_force)
                return forced_value
            except ValueError:
                return None
            except TypeError:
                return None

        # init variable
        values_in_list = False
        values_out_list = False
        value_is_dict = False

        if isinstance(field_value, (list, tuple, set)):
            value_origin_in_list = True
            values_in_list = True
            values_list = field_value
            if field_dict['field_width_list']:
                if len(str(field_value)) > field_dict['field_width_list']:
                    field_dict['field_width_list'] = len(str(field_value))
            else:
                field_dict['field_width_list'] = len(str(field_value))
        else:
            value_origin_in_list = False
            values_out_list = True
            if isinstance(field_value, dict):
                field_value = str(field_value)
                field_dict["field_recast"] = True
                value_is_dict = True
            values_list = [field_value]


        # loop on each value_to_force in list
        for value in values_list:
            # init var
            never_bool = False
            bool_type = None
            float_value = None
            # check if value is None
            if value in none_value_pattern:
                field_dict["none_value"] = True
                if value is not None:
                    field_dict["field_recast"] = True
                values_out_list = False
            # if value is not None
            else:
                field_dict["not_none_value"] = True
                # determine value_to_force type
                value_type = type(value)
                if value_is_dict is True:
                    field_dict['native_type'].update({dict})
                else:
                    field_dict['native_type'].update({value_type})
                test_hexadecimal = True
                # if try to force activate
                if field_try_to_force_type is True:
                    # we try to deduce if value_to_force is a list
                    iterable_value = False
                    if isinstance(value, str):
                        try:
                            eval_value = eval(value)
                            if isinstance(eval_value, (list, tuple, set)):
                                values_in_list = True
                                values_out_list = False
                                saving_native_type = set(field_dict['native_type'])
                                field_dict = define_field_type(
                                    field_dict=field_dict,
                                    field_value=eval_value,
                                    field_try_to_force_type=field_try_to_force_type
                                )
                                # restore native type
                                field_dict['native_type'] = saving_native_type
                                iterable_value = True

                            if isinstance(eval_value, bytes):
                                field_dict['tmp_field_type'].update({bytes})
                                # we do not need to test hexadecimal for this value
                                test_hexadecimal = False

                            # for this two new value_to_force recast is necessary
                            field_dict['field_recast'] = True

                        except SyntaxError:
                            pass
                        except ValueError:
                            pass
                        except NameError:
                            pass

                    # if not iterable we can deduce type
                    if not iterable_value:
                        bool_type = force_value(value, bool)
                        if never_bool is False:
                            if isinstance(bool_type, bool):
                                field_dict["tmp_field_type"].update({bool})
                            else:
                                if bool in field_dict["tmp_field_type"]:
                                    field_dict["tmp_field_type"].remove(bool)
                                never_bool = True

                        if not isinstance(value, bool):
                            # if not float try to force in float
                            if value_type != float:
                                float_value = force_value(value, float)
                                if float_value is None:
                                    field_dict["tmp_field_type"].update({float})
                            else:
                                float_value = value
                            # try to int
                            int_value = force_value(value, int)
                            if int_value is None:
                                field_dict["tmp_field_type"].update({int})
                            # if there is a difference between float and int value_to_force then value_to_force cannot
                            # be int
                            elif abs(float_value - int_value) != 0:
                                field_dict["tmp_field_type"].update({int})

                            # for bytes value
                            if value_type is bytes:
                                field_dict["tmp_field_type"].update({bytes})

                # if not try to force
                else:
                    # str
                    if isinstance(value, str):
                        field_dict["tmp_field_type"].update({str})
                    # float
                    elif isinstance(value, float):
                        field_dict["tmp_field_type"].update({float})
                    # int
                    elif isinstance(value, int):
                        # if bool
                        if isinstance(value, bool):
                            field_dict["tmp_field_type"].update({bool})
                        else:
                            field_dict["tmp_field_type"].update({int})

                    # bytes
                    elif isinstance(value, bytes):
                        field_dict["tmp_field_type"].update({bytes})
                    # iterable for iterable values in iterable values
                    if isinstance(value, (list, tuple, set)):
                        values_in_list = True
                        values_out_list = False
                        field_dict = define_field_type(field_dict=field_dict,
                                                       field_value=value,
                                                       field_try_to_force_type=field_try_to_force_type)

                # test if value is Date / Time or DateTime
                date_type = return_if_value_is_time_date_or_datetime(value=value, try_to_force=try_to_force_type)
                if date_type:
                    field_dict["tmp_field_type"].update({date_type})

                # define if str is always hexadecimal and can be converted to bytes
                if isinstance(value, str) and not date_type:
                    if (field_dict['str_is_always_hexadecimal'] is None or field_dict[
                        'str_is_always_hexadecimal'] is True) and test_hexadecimal is True:
                        field_dict['str_is_always_hexadecimal'] = is_hexadecimal(value) and (
                                len(value) % 2) == 0  # len of value must be even to be a bytes
                    if bool_type is None:
                        field_dict["tmp_field_type"].update({str})

                # Define width
                width_value = len(str(value))
                if value_origin_in_list is False and values_in_list is True:
                    pass  # here we have originaly a list in str that we have with force_value_type in True convert in
                    # list : we don't calculate field_width_str at this step.
                else:
                    if width_value > field_dict["field_width_str"]:
                        field_dict["field_width_str"] = width_value

                # Define precision and modify with if necessary
                if field_try_to_force_type is False:
                    float_value = value
                if isinstance(float_value, float):
                    value_split = str(value).split(".")
                    if len(value_split) == 2:
                        before_comma_value, after_comma_values = value_split
                    else:
                        before_comma_value = value_split[-1]
                        after_comma_values = '0'

                    if after_comma_values == '0':
                        width_after_comma = 0
                    else:
                        width_after_comma = len(after_comma_values)

                    if before_comma_value == '0':
                        width_before_comma = 0
                    else:
                        width_before_comma = len(before_comma_value)

                    if width_after_comma > field_dict["field_precision"]:
                        field_dict["field_precision"] = width_after_comma
                        field_dict["width_after_comma"] = width_after_comma

                    if width_before_comma > field_dict["width_before_comma"]:
                        field_dict["width_before_comma"] = width_before_comma

                    # update width
                    if field_dict["width_before_comma"] + field_dict["width_after_comma"] > field_dict[
                        "field_width_float"]:
                        field_dict["field_width_float"] = field_dict["width_before_comma"] + field_dict[
                            "width_after_comma"]

            # if list or not
            if values_in_list is True and field_dict["values_in_list"] is False:
                field_dict["values_in_list"] = True

            if values_out_list is True and field_dict["values_out_list"] is False:
                field_dict["values_out_list"] = True

        return field_dict

    def loop_on_each_feature(
            _geolayer_or_feature_list,
            _is_geolayer,
            _scan_all_field_name_for_each_feature,
            _field_name_list,
            _try_to_force_type
    ):
        """
        Loop on each feature (in feature_list or geolayer).

        :param _geolayer_or_feature_list: list of all features or geolayer
        :param _is_geolayer: True if _feature_list is geolayer / False if feature_list.
        :param _scan_all_field_name_for_each_feature: is True all field name in features will be scan.
        :param _field_name_list: field name list that we want to scan on geolayer or feature list.
        :param _try_to_force_type: option that can force the Type of value on geolayer or features list.
        :return: result from fields scan of each features
        """

        # init output values
        fields_scan = {}
        field_index = 0
        fields_name_set = set()

        # loop on each features
        for i, feature in enumerate(_geolayer_or_feature_list):
            if _is_geolayer:
                feature = geolayer_or_feature_list["features"][feature]
            if "attributes" in feature:
                # if we loop on each field in feature, we get all fields in feature
                if _scan_all_field_name_for_each_feature:
                    _field_name_list = feature["attributes"].keys()

                # we loop on feature's fields
                # if field does not exist in fields_name_set we create it enter in fields_scan
                for i_field, field_name_filter in enumerate(_field_name_list):
                    # first apparition initialise dico
                    if field_name_filter not in fields_name_set:
                        fields_name_set.update({field_name_filter})
                        fields_scan[field_name_filter] = {
                            "values_in_list": False,
                            "values_out_list": False,
                            "field_list": False,
                            "tmp_field_type": set(),
                            "field_type": None,
                            "field_width_str": 0,
                            "field_width_list": None,
                            "field_precision": 0,
                            "none_value": False,
                            "not_none_value": False,
                            "field_index": field_index,
                            "field_recast": False,
                            "native_type": set(),
                            "force_type": False,
                            "str_is_always_hexadecimal": None,
                            "field_width_float": 0,
                            "width_before_comma": 0,
                            "width_after_comma": 0,
                        }
                        # if we create a new field_name in dict not at first feature that means
                        # that for the previous entities the value of the field was None.
                        if i != 0:
                            fields_scan[field_name_filter]['none_value'] = True
                        if _try_to_force_type is True:
                            fields_scan[field_name_filter]['force_type'] = True
                        # save field field_index
                        field_index += 1

                    # if field_name is in feature we deduce type of field value
                    if field_name_filter in feature["attributes"]:
                        feature_field_value = feature["attributes"][field_name_filter]
                        fields_scan[field_name_filter] = define_field_type(field_dict=fields_scan[field_name_filter],
                                                                           field_value=feature_field_value,
                                                                           field_try_to_force_type=_try_to_force_type)
                    else:
                        fields_scan[field_name_filter]['none_value'] = True

                # test if missing field name
                if _scan_all_field_name_for_each_feature:
                    missing_field_name_set = (fields_name_set.difference(_field_name_list))
                    if missing_field_name_set:
                        for missing_field_name in missing_field_name_set:
                            fields_scan[missing_field_name]['none_value'] = True

        return fields_scan

    def deduce_fields_type_from_raw_metadata(_feature_list_fields_scan, _try_to_force_type):
        """
        From raw metadata result deduce field type.

        :param _feature_list_fields_scan: raw metadata dict
        :param _try_to_force_type: True if you want force type value
        :return: raw metadata dict with feature type in it.
        """
        # deduce output field type
        for _field_name_filter in _feature_list_fields_scan:
            feature_tmp_field_type = _feature_list_fields_scan[_field_name_filter].get('tmp_field_type', None)
            if _try_to_force_type is True:
                # WARNING
                # When _try_to_force_type is TRUE
                # types stored in 'tmp_field_type' can contains types that are NOT compatible with data thus we
                # can DEDUCE the type.
                if feature_tmp_field_type:
                    str_force = True
                    field_type_set = {}
                    # numeric fields (float and int)
                    if float not in feature_tmp_field_type and int in \
                            feature_tmp_field_type:
                        field_type_set = {float}
                        str_force = False

                    if int not in feature_tmp_field_type:
                        field_type_set = {int}
                        _feature_list_fields_scan[_field_name_filter]['field_precision'] = 0
                        str_force = False

                    # bytes
                    if bytes in feature_tmp_field_type and float in \
                            _feature_list_fields_scan[_field_name_filter][
                                'tmp_field_type'] and int in _feature_list_fields_scan[_field_name_filter][
                        'tmp_field_type']:
                        # if there is other type in 'tmp_field_type' it must be str else we force to str
                        if feature_tmp_field_type - {int, float, bytes} == {
                            str} or feature_tmp_field_type - {int, float,
                                                              bytes} == set():
                            # if str is always hexadecimal then it's a bytes valid
                            if str in feature_tmp_field_type and \
                                    _feature_list_fields_scan[_field_name_filter]['str_is_always_hexadecimal'] is False:
                                str_force = True
                            else:
                                field_type_set = {bytes}
                                str_force = False
                        else:
                            str_force = True
                    # bytes 2 (if data is only hexadecimal but not integer only)
                    if feature_tmp_field_type - {float, str, int} == set() and \
                            _feature_list_fields_scan[_field_name_filter][
                                'str_is_always_hexadecimal'] is True and int not in field_type_set:
                        field_type_set = {bytes}
                        str_force = False

                    # datetime / date and time and bool
                    # create temp_field_type variable
                    if {bool, datetime.date, datetime.time, datetime.datetime}.intersection(feature_tmp_field_type):
                        date_time_datetime_and_bool = feature_tmp_field_type - {int, float}
                        temp_field_type_for_date_time_datetime_and_bool = date_time_datetime_and_bool.difference(
                            {bool, datetime.date, datetime.time, datetime.datetime})
                        if len(temp_field_type_for_date_time_datetime_and_bool) == 0:
                            # if one type
                            if len(date_time_datetime_and_bool) == 1:
                                field_type_set = date_time_datetime_and_bool
                                str_force = False
                            else:
                                # if bool we must recast field to str
                                if bool in date_time_datetime_and_bool:
                                    field_type_set = date_time_datetime_and_bool
                                    str_force = True
                                else:
                                    # if time / date and datetime (no compatibility we force to str)
                                    if len(date_time_datetime_and_bool) == 3:
                                        field_type_set = date_time_datetime_and_bool
                                        str_force = True
                                    else:
                                        if date_time_datetime_and_bool == {datetime.datetime, datetime.date}:
                                            field_type_set = {datetime.datetime}
                                            str_force = False
                                        else:
                                            field_type_set = date_time_datetime_and_bool
                                            str_force = True

                    # define if value in list must be forced to str
                    if _feature_list_fields_scan[_field_name_filter]['values_in_list'] is True and \
                            _feature_list_fields_scan[_field_name_filter]['values_out_list'] is True:
                        str_force = True

                    # force in str
                    if str_force is True:
                        field_type_set = {str}

                    feature_tmp_field_type = field_type_set

                else:
                    # for  integer value
                    if _feature_list_fields_scan[_field_name_filter]['not_none_value'] is True:
                        feature_tmp_field_type = {int}
                        _feature_list_fields_scan[_field_name_filter]['field_precision'] = 0

            # if try_to_force is False
            else:
                if len(feature_tmp_field_type) > 1:
                    _feature_list_fields_scan[_field_name_filter]['field_recast'] = True

                    if str in feature_tmp_field_type:
                        feature_tmp_field_type = {str}
                    else:
                        datetime_set = {datetime.date, datetime.time, datetime.datetime}
                        # boolean or integer
                        if bool in feature_tmp_field_type and int in feature_tmp_field_type:
                            feature_tmp_field_type = {str}
                        # float
                        elif float in feature_tmp_field_type and int in feature_tmp_field_type and len(
                                feature_tmp_field_type) == 2:
                            feature_tmp_field_type = {float}

                        # date / time / datetime
                        elif len(feature_tmp_field_type.difference(datetime_set)) == 0:
                            date_time_datetime_and_bool = feature_tmp_field_type.intersection(datetime_set)
                            # if time / date and datetime (no compatibility we force to str)
                            if len(date_time_datetime_and_bool) == 3:
                                datetime_field_type = {str}
                            else:
                                if date_time_datetime_and_bool == {datetime.datetime, datetime.date}:
                                    datetime_field_type = {datetime.datetime}
                                else:
                                    datetime_field_type = {str}
                            feature_tmp_field_type = datetime_field_type
                        else:
                            feature_tmp_field_type = {str}

            _feature_list_fields_scan[_field_name_filter]['tmp_field_type'] = feature_tmp_field_type

        # Rescan all fields data and determine :
        # - deduce final type of data in field and delete field with only None value
        # - are on a list (only for str / int / float data)

        # write final type field and delete field with only None value ('reindex other field is necessary')
        for _field_name_filter, field_name_dict in _feature_list_fields_scan.items():
            # determine field type
            if field_name_dict['tmp_field_type']:
                field_type = list(field_name_dict['tmp_field_type'])[0]
                field_name_dict['field_type'] = field_type
            else:
                # for none type field
                field_name_dict['field_delete'] = True

        # list
        for _field_name_filter, field_name_dict in _feature_list_fields_scan.items():
            type_list = False

            if field_name_dict['field_type'] in {str, int, float}:
                if field_name_dict['values_in_list'] is True and field_name_dict['values_out_list'] is False:
                    type_list = True

            field_name_dict['field_list'] = type_list

        return _feature_list_fields_scan

    def check_field_to_delete(_feature_list_fields_scan):
        """
        Scan field scan result : add key 'field_delete' and deduce if field must be deleted or not

        :param _feature_list_fields_scan: fields scan dict
        :return: _feature_list_fields_scan with 'field_delete' key
        """
        for _field_name_filter, _field_name_dict in _feature_list_fields_scan.items():
            _field_name_dict['field_delete'] = False
            if _field_name_dict['none_value'] is True and _field_name_dict['not_none_value'] is False:
                _field_name_dict['field_delete'] = True

        return _feature_list_fields_scan

    def check_field_to_recast(_feature_list_fields_scan):
        """
        Scan field scan result : update key 'field_recast' and deduce if field must be recast or not

        :param _feature_list_fields_scan: fields scan dict
        :return:  _feature_list_fields_scan with 'field_recast' updated
        """

        # determine if field must be recast
        for _field_name_filter, _field_name_dict in _feature_list_fields_scan.items():
            if len(_field_name_dict['native_type']) > 1 or \
                    _field_name_dict['native_type'] != _field_name_dict['tmp_field_type']:
                _field_name_dict['field_recast'] = True
            if _field_name_dict['field_type'] in {str, int, float}:
                if _field_name_dict['values_in_list'] is True and _field_name_dict['values_out_list'] is True:
                    _field_name_dict['field_recast'] = True
            else:
                if _field_name_dict['values_in_list'] is True:
                    raise Exception("field {field_name} : field_type must be change to str".format(
                        field_name=_field_name_filter))

        return _feature_list_fields_scan

    # init input values
    # check if input data is geolayer or feature
    if isinstance(geolayer_or_feature_list, list):
        is_geolayer = False
        feature_list = geolayer_or_feature_list
    elif isinstance(geolayer_or_feature_list, dict):
        is_geolayer = True
        feature_list = geolayer_or_feature_list["features"]
    else:
        raise Exception('geolayer_or_feature_list must be a list of features or a Geolayer')

    scan_all_field_name_for_each_feature = True
    if field_name_filter:
        scan_all_field_name_for_each_feature = False
        field_name_list = value_to_iterable_value(field_name_filter, list)
    else:
        field_name_list = None

    # loop on each feature and get raw metadata
    feature_list_fields_scan = loop_on_each_feature(
        _geolayer_or_feature_list=feature_list,
        _is_geolayer=is_geolayer,
        _scan_all_field_name_for_each_feature=scan_all_field_name_for_each_feature,
        _field_name_list=field_name_list,
        _try_to_force_type=try_to_force_type
    )
    # from raw metadata deduce field type
    feature_list_fields_scan = deduce_fields_type_from_raw_metadata(
        _feature_list_fields_scan=feature_list_fields_scan,
        _try_to_force_type=try_to_force_type
    )

    # check if field(s) must be delete
    feature_list_fields_scan = check_field_to_delete(_feature_list_fields_scan=feature_list_fields_scan)
    # check if field(s) must be recast or not
    feature_list_fields_scan = check_field_to_recast(_feature_list_fields_scan=feature_list_fields_scan)

    # clean output
    for field_name_filter, field_name_dict in feature_list_fields_scan.items():
        del field_name_dict['tmp_field_type']
        if fields_index is False:
            del field_name_dict['field_index']

    return feature_list_fields_scan


def feature_filter_geometry(
        feature,
        geometry_type_filter=None,
        bbox_filter=None,
        bbox=True
):
    """
    return a geometry :
        - a certain type of geometries when geometry_type_filter is filled
        - a geometry that intersect a given bbox if bbox_filter is filled

    :param feature: feature that will be filtered
    :param geometry_type_filter: Geometry(ies) type(s) that we want to keep in feature.
        If type not existing the feature is None.
    :param bbox_filter: if bbox(s) intersects features return feature. If feature not intersecting bbox function return
        None
    :param bbox:
    :return: geometry part of feature (beware this is not a feature on output but only geometry part)
    """
    if feature:
        feature = copy.deepcopy(feature)
        # TODO remove when bbox will be an object
        if isinstance(bbox_filter, (list, tuple)):
            if isinstance(bbox_filter[0], (int, float)):
                bbox_filter = [bbox_filter]
        bbox_filter = value_to_iterable_value(bbox_filter, tuple)

        geometry = {}
        if 'geometry' in feature:
            if geometry_type_filter:
                # convert geometry to geometry collection (we use geometry_type_filter to filter geometries)
                geometry_in_collection = geometry_to_geometry_collection(
                    geometry=feature['geometry'],
                    geometry_type_filter=geometry_type_filter,
                    bbox=False
                )
                if geometry_in_collection['geometries']:
                    # if we get only one geometry in geometryCollection we return the only geometry in it
                    if len(geometry_in_collection['geometries']) == 1:
                        geometry = geometry_in_collection['geometries'][0]
                    else:
                        geometry = geometry_in_collection
                else:
                    geometry = {}
            else:
                geometry = feature['geometry']

            if bbox_filter and geometry:
                geometry_in_bbox = False
                geometry_bbox = geometry_to_bbox(geometry)
                if geometry_bbox:
                    for bbox_in_filter in bbox_filter:
                        if bbox_intersects_bbox(geometry_bbox, bbox_in_filter):
                            geometry_in_bbox = True
                            break

                if geometry_in_bbox is False:
                    geometry = {}

            # if bbox option is activate we compute it
            if bbox is True and geometry:
                if 'bbox' not in feature['geometry']:
                    if geometry['type'] == 'GeometryCollection':
                        for geometry_in_collection in geometry['geometries']:
                            geometry_in_collection_bbox = geometry_to_bbox(geometry_in_collection)
                            if geometry_in_collection_bbox:
                                geometry_in_collection['bbox'] = geometry_in_collection_bbox

                    geometry_bbox = geometry_to_bbox(geometry)
                    if geometry_bbox:
                        geometry['bbox'] = geometry_bbox
        return geometry


def feature_filter_attributes(feature, field_name_filter=None):
    """
    Keeps (filter) only the fields specified in the variable field_name_filter

    :param feature: feature that will be filtered
    :param field_name_filter: field name that we want to keep in feature (if present in feature).
    :return: attributes part of feature (beware this is not a feature on output but only attributes part)
    """
    # initialize input
    field_name_filter = value_to_iterable_value(field_name_filter, list)

    if feature:
        # initialize output
        new_feature_attributes = {}

        if field_name_filter:

            # format field_name_filter
            field_name_filter = value_to_iterable_value(field_name_filter)

            if 'attributes' in feature:
                if field_name_filter:
                    for field_name in field_name_filter:
                        if field_name in feature["attributes"]:
                            if new_feature_attributes:
                                new_feature_attributes[field_name] = feature['attributes'][field_name]
                            else:
                                new_feature_attributes = {field_name: feature['attributes'][field_name]}
        else:
            if 'attributes' in feature:
                new_feature_attributes = feature['attributes']

        return new_feature_attributes


def feature_filter(
    feature,
    serialize=None,
    field_name_filter=None,
    geometry_type_filter=None,
    bbox_filter=None,
    bbox=False
):
    """
    This function apply filter on "attributes" and/or "geometry"
        Attributes filter
            - field name filter

        Geometry filter
            - geometry type filter
            - bbox filter if feature geometry


    :param serialize: if you want to serialize your data
    :param feature: feature that we want filter
    :param field_name_filter: field name that we want to keep in feature (if present in feature).
    :param geometry_type_filter: Geometry(ies) type(s) that we want to keep in feature.
        If type not existing the feature is None.
    :param bbox_filter: if bbox(s) intersects features return feature. If feature not intersecting bbox function return
        None
    :param bbox: if you want to compute feature geometry bbox
    :return: filtered feature
    """
    # check value to deduce if we must filter feature or not
    if serialize or field_name_filter or geometry_type_filter or bbox_filter or bbox:
        # initialize input variable
        # attributes
        attributes_filter = False
        if field_name_filter:
            attributes_filter = True
        # geometry
        geometry_filter = False
        if geometry_type_filter or bbox_filter:
            geometry_filter = True

        # initialize output variable
        new_feature = {}
        # attributes filter
        feature_attributes = feature_filter_attributes(
            feature,
            field_name_filter=field_name_filter
        )
        if feature_attributes:
            new_feature['attributes'] = feature_attributes

        # geometry filter
        feature_geometry = feature_filter_geometry(
            feature,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            bbox=bbox
        )
        if feature_geometry:
            new_feature['geometry'] = feature_geometry

        # check if feature is valid
        if attributes_filter and 'attributes' not in new_feature and 'geometry' not in new_feature:
            new_feature = None

        if geometry_filter and new_feature and 'geometry' not in new_feature:
            new_feature = None

        if new_feature:
            if serialize:
                new_feature = feature_serialize(new_feature)

            return new_feature

    else:
        return feature


def features_filter(
        geolayer_feature_list_or_generator,
        field_name_filter=None,
        geometry_type_filter=None,
        bbox_filter=None,
        serialize=None,
        bbox_extent=True,
        feature_limit=None,
        feature_offset=None
):
    # if input is a geolayer
    if isinstance(geolayer_feature_list_or_generator, dict):
        feature_i_feat_list, geolayer_feature_list_or_generator = zip(*geolayer_feature_list_or_generator['features'])

    yield_count = 0
    for i_feat, feature in enumerate(geolayer_feature_list_or_generator):
        if feature_offset:
            if i_feat < feature_offset:
                continue

        output_feature = feature_filter(
            feature=feature,
            serialize=serialize,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            bbox=bbox_extent,
        )
        if output_feature:
            yield output_feature
            yield_count += 1

            if feature_limit:
                if yield_count == feature_limit:
                    break





from geoformat.conversion.geometry_conversion import geometry_to_bbox


def merge_geometries(geom_a, geom_b, bbox=True):
    """
    Return a merging geometry result of adding two different geometries
    Be careful this function does not "union" two geometrie that intersects. It adds a geometry to another

    Merging Table :

        Single AND Single
            Point + Point = MultiPoint
            LineString + LineString = MultiLineString
            Polygon + Polygon = MultiPolygon

        Single AND Multi
            Point + MultiPoint = MultiPoint
            LineString  + MultiLineString = MultiLineString
            Polygon + MultiPolygon = MultiPolygon

        Mixed Geometries Types and GeometryCollection
            Point + Polygon = GeometryCollection(Point, Polygon)
            GeometryCollection(Polygon + LineString) + LineSting = GeometryCollection(Polygon + MultiLineString)
            GeometryCollection(MultiPolygon, LineString), GeometryCollection(MultiPoint, LineString)
                = GeometryCollection(MultiPolygon, MultiLineString, MultiPoint)

    How does it work ?

        - first if geometry categories are the same
        - if geometry categories are different or GeometryCollection it's return a GeometryCollection

    :param geom_a: first geometry to merge.
    :param geom_b: second geometry to merge.
    :param bbox: True : add bbox to geometry / False no bbox on geometry.
    """
    new_geom = {}

    # check if geometries are empty
    geom_a_not_empty, geom_b_not_empty = False, False

    if geom_a.get('coordinates', geom_a.get('geometries')):
        geom_a_not_empty = True
    if geom_b.get('coordinates', geom_b.get('geometries')):
        geom_b_not_empty = True

    # if geometry is not empty
    if geom_a_not_empty is True and geom_b_not_empty is True:
        # if same geometry category (Point, MultiPoint), (LineString, MultiLineString), (Polygon, MultiPolygon)
        if geom_a['type'].replace('Multi', '') == geom_b['type'].replace('Multi', '') and geom_a[
            'type'] != 'GeometryCollection' and geom_b['type'] != 'GeometryCollection':
            new_geom_type = str(geom_a['type'])
            coordinates_or_geometries = list(geom_a['coordinates'])
            key_coordinates_or_geometry = 'coordinates'
            if 'Multi' in geom_a['type']:
                if 'Multi' in geom_b['type']:
                    for geom_coordinates in geom_b['coordinates']:
                        if geom_coordinates not in coordinates_or_geometries:
                            coordinates_or_geometries.append(list(geom_coordinates))
                else:
                    if geom_b['coordinates'] not in coordinates_or_geometries:
                        coordinates_or_geometries.append(list(geom_b['coordinates']))

            else:
                if 'Multi' in geom_b['type']:
                    coordinates_or_geometries = [coordinates_or_geometries]
                    len_coordinates_or_geometries = len(coordinates_or_geometries)
                    for geom_coordinates in geom_b['coordinates']:
                        if geom_coordinates not in coordinates_or_geometries:
                            coordinates_or_geometries.append(list(geom_coordinates))
                    if len_coordinates_or_geometries < len(coordinates_or_geometries):
                        new_geom_type = 'Multi' + new_geom_type
                else:
                    if geom_a['coordinates'] != geom_b['coordinates']:
                        coordinates_or_geometries = [coordinates_or_geometries, list(geom_b['coordinates'])]
                        # add Mutli to type
                        new_geom_type = 'Multi' + new_geom_type

            new_geom['type'] = new_geom_type
            new_geom[key_coordinates_or_geometry] = coordinates_or_geometries
        else:
            if geom_a['type'] == 'GeometryCollection' and geom_b['type'] == 'GeometryCollection':
                # first loop on geom_a geometries
                for i_a, geojson_geometrie_a in enumerate(geom_a['geometries']):
                    if i_a == 0:
                        new_geom = geojson_geometrie_a
                    else:
                        new_geom = merge_geometries(new_geom, geojson_geometrie_a, bbox=bbox)

                # add geom_ b geometries
                for geojson_geometrie_b in geom_b['geometries']:
                    new_geom = merge_geometries(new_geom, geojson_geometrie_b, bbox=bbox)

            elif geom_a['type'] == 'GeometryCollection' or geom_b['type'] == 'GeometryCollection':
                if geom_a['type'] == 'GeometryCollection':
                    ori_geom_collect = dict(geom_a)
                    ori_geom_simple = dict(geom_b)
                else:
                    ori_geom_collect = dict(geom_b)
                    ori_geom_simple = dict(geom_a)

                # first loop on ori_geom_collect geometries
                for i_a, geojson_geometrie_a in enumerate(ori_geom_collect['geometries']):
                    if i_a == 0:
                        new_geom = geojson_geometrie_a
                    else:
                        new_geom = merge_geometries(new_geom, geojson_geometrie_a, bbox=bbox)

                added_geom = False
                # then we see if ori_geom_simple as similar geom type in ori_geom_collect else we had
                for i_geom, geometry in enumerate(new_geom['geometries']):
                    if geometry['type'].replace('Multi', '') == ori_geom_simple['type'].replace('Multi', ''):
                        replace_geom = merge_geometries(geometry, ori_geom_simple, bbox=bbox)
                        new_geom['geometries'][i_geom] = replace_geom
                        added_geom = True
                        # end loop
                        break

                if added_geom is False:
                    # if not break we had ori_geom_simple to new_geom GEOMETRYCOLLECTION
                    new_geom['geometries'] = new_geom['geometries'] + [dict(ori_geom_simple)]

            else:
                new_geom['type'] = 'GeometryCollection'
                new_geom['geometries'] = [dict(geom_a), dict(geom_b)]

    else:
        if geom_a_not_empty is True:
            new_geom_type = str(geom_a['type'])
            coordinates_or_geometries = list(geom_a['coordinates'])
        elif geom_b_not_empty is True:
            new_geom_type = str(geom_b['type'])
            coordinates_or_geometries = list(geom_b['coordinates'])

        # both geometries are empty
        else:
            if geom_a['type'] == geom_b['type']:
                new_geom_type = str(geom_a['type'])
            else:
                new_geom_type = 'GeometryCollection'

            coordinates_or_geometries = []

        if new_geom_type == 'GeometryCollection':
            key_coordinates_or_geometry = 'geometries'
        else:
            key_coordinates_or_geometry = 'coordinates'

        # add type and coordinates/geometries
        new_geom['type'] = new_geom_type
        new_geom[key_coordinates_or_geometry] = coordinates_or_geometries

    if bbox is True:
        new_geom['bbox'] = geometry_to_bbox(new_geom)

    return new_geom



from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance_point_vs_segment
)


def ramer_douglas_peucker(coordinate_list, tolerance):
    """
    This function is a generalizatio algorithm.
    It's take a coordinate's list and a tolerance value and return a simplified coordinate's list

        https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    :param coordinate_list: coordinate's list
    :param tolerance: maximum distance in coordinates units
    :return: simplified coordinates
    """

    # if coordinates_list is not empty and have more than 2 coordinates
    if len(coordinate_list) > 2:
        # get coordinates_list extremity points
        point_start = coordinate_list[0]
        point_end = coordinate_list[-1]

        # create segment with extremity points
        segment = [point_start, point_end]

        # initialize coordinate variables
        distance_max_point_segment = 0
        index_max_point_segment = None
        coordinate_list_without_start_end_points = coordinate_list[1:-1]

        for i_point, point in enumerate(coordinate_list_without_start_end_points):
            distance_point_segment = euclidean_distance_point_vs_segment(point, segment)
            if distance_point_segment > tolerance:
                if distance_point_segment > distance_max_point_segment:
                    distance_max_point_segment = distance_point_segment
                    index_max_point_segment = i_point

        if index_max_point_segment is None:
            coordinate_list_result = segment
        else:
            coordinate_list_start = [point_start] + coordinate_list_without_start_end_points[:index_max_point_segment+1]
            coordinate_list_end = coordinate_list_without_start_end_points[index_max_point_segment:] + [point_end]
            segment_start = ramer_douglas_peucker(coordinate_list=coordinate_list_start, tolerance=tolerance)
            segment_end = ramer_douglas_peucker(coordinate_list=coordinate_list_end, tolerance=tolerance)
            coordinate_list_result = segment_start[:-1] + segment_end
    else:
        coordinate_list_result = coordinate_list

    return coordinate_list_result




from geoformat.index.geometry.grid import (
    create_grid_index,
    bbox_to_g_id
)
from geoformat.conversion.geometry_conversion import geometry_to_ogr_geometry
from geoformat.geoprocessing.connectors.operations import coordinates_to_bbox

########################################################################################################################
#
# Adjacency matrix
#
########################################################################################################################


def create_adjacency_matrix(geolayer, mesh_size=None):
    """
    Creating an adjacency matrix.

    Return a dict with this structure :
    {
        "metadata": {
            "type": "adjacency",
        }
        "matrix": {i_feat_foo:  set(i_feat_bar), i_feat_bar: set(i_feat_foo), i_feat_baz : set()}
    }

    * GDAL/OGR dependencie :
        Intersects
    *
    """

    matrix_dict = {}

    try:
        input_grid_idx = geolayer['metadata']['constraints']['index']['geometry']
    except:
        input_grid_idx = create_grid_index(geolayer, mesh_size=mesh_size)

    mesh_size = input_grid_idx['metadata']['mesh_size']

    # Création de la clé unique du dictionnaire de résultat (préparation du stockage des résultats)
    matrix_dict['matrix'] = {i_feat: set() for i_feat in geolayer['features']}

    for i_feat in geolayer['features']:
        feature_a = geolayer['features'][i_feat]

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature_a = eval(feature_a)

        # get bbox
        try:
            feature_a_bbox = feature_a['geometry']['bbox']
        except KeyError:
            feature_a_bbox = coordinates_to_bbox(feature_a['geometry']['coordinates'])
            feature_a['geometry']['bbox'] = feature_a_bbox

        # Récupération des identifiants de mailles présent dans l'index
        g_id_list = list(bbox_to_g_id(feature_a_bbox, mesh_size))

        # Pour chaque identifiant de maille de l'index on récupère l'identifant des autres polygones voisins de l'entité
        # en cours (s'ils existent)
        neighbour_i_feat_list = []
        for g_id in g_id_list:
            # récupération de la liste des clefs primaires des entités présentes dans la maille de l'index
            neighbour_i_feat_list += input_grid_idx['index'][g_id]

        # suppression de la clef primaire du polyone en cours et d'éventuels doublons
        neighbour_i_feat_list = [value for value in set(neighbour_i_feat_list) if value != i_feat]

        # création  de la géométrie du feature A
        feature_a_ogr_geom = geometry_to_ogr_geometry(feature_a['geometry'])
        # Même procédé que précédemment pour l'objet B
        for neighbour_i_feat in neighbour_i_feat_list:
            # si le i_feat n'est pas déjà compris dans la matrice de proximité du voisin :
            #  si c'est le cas alors cela veut dire que l'on a déjà fait un test d'intersection enrte les deux
            #  géométries auparavant (car pour avoir un voisin il faut etre deux)

            if i_feat not in matrix_dict['matrix'][neighbour_i_feat]:
                feature_b = geolayer['features'][neighbour_i_feat]

                # if feature is serialized
                if 'feature_serialize' in geolayer['metadata']:
                    if geolayer['metadata']['feature_serialize']:
                        feature_b = eval(feature_b)

                feature_b_ogr_geom = geometry_to_ogr_geometry(feature_b['geometry'])

                if feature_a_ogr_geom.Intersects(feature_b_ogr_geom):
                    matrix_dict['matrix'][i_feat].add(neighbour_i_feat)
                    matrix_dict['matrix'][neighbour_i_feat].add(i_feat)

    matrix_dict['metadata'] = {'type': 'adjacency'}

    return matrix_dict


def get_area_intersecting_neighbors_i_feat(adjacency_matrix, i_feat, neighbor_set=None):
    """
    Use adjacency matrix to find all i_feat neighbor's.
        Return all geometries that intersects and intersects neihgbor

    """

    if not neighbor_set:
        neighbor_set = set([i_feat])

    # set store i_feat neighbor's
    i_feat_neighbor_set = set(adjacency_matrix['matrix'][i_feat])
    # difference result between the list of neighbors already scanned and i_feat neighbors
    # Anyway, it's the list of new neighbors we haven't scanned yet.
    new_neighbor_set = set(i_feat_neighbor_set.difference(neighbor_set))
    # copy previous set
    new_neighbor_set_copy = set(new_neighbor_set)

    i = 0
    while new_neighbor_set_copy:
        # Loop scan every new neighbor
        for neighbor in new_neighbor_set:
            # set store i_feat neighbor's
            i_feat_neighbor_set = set(adjacency_matrix['matrix'][neighbor])
            # new set of neighbors who have never been scanned before
            new_new_neighbor_set = i_feat_neighbor_set.difference(neighbor_set)
            # add the neighbor to the result list
            neighbor_set.update({neighbor})
            # we add the new neighbors never scanned before
            new_neighbor_set_copy.update(new_new_neighbor_set)

            # we delete the scanned neighbor
            new_neighbor_set_copy = new_neighbor_set_copy.difference({neighbor})

            i += 1
        # adding to new_neighbor_set ew neighbors that have never been scanned before
        new_neighbor_set = set(new_neighbor_set_copy)

    return neighbor_set


def get_neighbor_i_feat(adjacency_matrix, i_feat):
    """
    Return neighbor for given i_feat

    :return:
    """

    return adjacency_matrix["matrix"][i_feat]



import json

from geoformat.conf.fields_variable import geoformat_field_type_to_geojson_field_type
from geoformat.conf.format_data import value_to_iterable_value
from geoformat.conf.geoformat_var import GEOLAYER_DEFAULT_NAME
from geoformat.conf.geometry_variable import GEOFORMAT_GEOMETRY_TYPE
from geoformat.conf.path import path_to_file_path
from geoformat.conf.proj_var import proj_correspondance
from geoformat.conversion.feature_conversion import feature_deserialize
from geoformat.conversion.feature_conversion import features_filter
from geoformat.conversion.fields_conversion import recast_field_value
from geoformat.driver.common_driver import _get_recast_field_type_mapping, load_data
from geoformat.manipulation.geolayer_manipulation import feature_list_to_geolayer


def feature_attributes_to_properties(attributes, recast_field_mapping=None):
    """
    Transforms feature's attributes to geojson properties

    :param attributes: feature attributes
    :param recast_field_mapping: config dict to recast field to geojson properties format
    :return: formatted geojson properties
    """
    if attributes:
        # check is data are serialized or not
        properties = dict(attributes)

        # if recast_field_mapping
        if recast_field_mapping is not None:
            for field_name, field_name_mapping in recast_field_mapping.items():
                field_value_to_recast = attributes.get(field_name, None)
                if field_value_to_recast:
                    field_value_recast = recast_field_value(field_value=field_value_to_recast, **field_name_mapping)
                    properties[field_name] = field_value_recast
    else:
        properties = {}

    return properties


def geoformat_feature_to_geojson_feature(feature, i_feat=None, recast_field_mapping=None):
    """
    Make transformation between geoformat feature to geojson feature.

    :param feature: geoformat feature
    :param i_feat: geoformat feature's id
    :param recast_field_mapping:
    :return: dict like geojson feature
    """
    # get geometries
    geometry = dict(feature.get('geometry', {}))
    # delete bbox if exists
    geometry.pop('bbox', None)
    # get attributes and clean it
    attributes = dict(feature.get('attributes', {}))
    properties = feature_attributes_to_properties(attributes, recast_field_mapping)
    geojson_feature = dict(
        type='Feature',
        geometry=geometry,
        properties=properties
    )
    if i_feat is not None:
        geojson_feature['id'] = i_feat

    return geojson_feature




def geolayer_to_geojson(
    geolayer,
    path,
    overwrite=False,
    add_extension=False,
    indent=4,
    encoding='utf8'
):
    """
    Save geolayer to geojson format

    :param geolayer: input geolayer.
    :param path: output geojson file or directory path .
    :param overwrite: True we overwrite if exists false we raise an exception.
    :param add_extension: True we add .geojson extension to file path if extension .geojson not exists.
    :param indent: format output file with specific indentation.
    :param encoding: string encoding default utf8.

    """

    if add_extension:
        add_extension = '.geojson'

    # check input path and deduce output path
    output_path = path_to_file_path(
        path=path,
        geolayer_name=geolayer['metadata']['name'],
        overwrite=overwrite,
        add_extension=add_extension
    )

    # scan field metadata to check if field must have to be recast
    fields_metadata = geolayer['metadata'].get('fields', {})
    recast_field_mapping = _get_recast_field_type_mapping(
        fields_metadata=fields_metadata,
        geoformat_type_to_output_driver_type=geoformat_field_type_to_geojson_field_type
    )

    geojson_features_list = [None] * len(geolayer['features'])
    for i, (i_feat, feature) in enumerate(geolayer['features'].items()):
        if geolayer['metadata'].get('serialize', False):
            feature = feature_deserialize(feature)
        geojson_feature = geoformat_feature_to_geojson_feature(
            feature=feature,
            i_feat=i_feat,
            recast_field_mapping=recast_field_mapping)
        geojson_features_list[i] = geojson_feature

    # create FeatureCollection
    name = geolayer['metadata']['name']
    geojson_feature_collection = {'type': "FeatureCollection", "name": name, "features": geojson_features_list}

    # if reference coordinate system (crs) is filled
    if 'geometry_ref' in geolayer['metadata']:
        if 'crs' in geolayer['metadata']['geometry_ref']:
            geolayer_crs = geolayer['metadata']['geometry_ref']['crs']
            crs_string = {
                "type": "name",
                "properties": {
                    "name": "urn:ogc:def:crs:EPSG::{crs}".format(crs=geolayer_crs)
                }
            }
            crs_string = crs_string
            geojson_feature_collection['crs'] = crs_string

    # write geojson file
    with open(output_path, 'w', encoding=encoding) as geojson_file:
        json.dump(obj=geojson_feature_collection, fp=geojson_file, indent=indent)


# def load_json_data(path, http_headers=None, encoding=None):
#     """
#     Return a tuple that contain : json load to str (index 0) and name of geojson (index 1)
#
#     :param path: json path
#     :param http_headers: for http request we can use optionally headers
#     :return: json load to str (index 0) and name of geojson (index 1)
#     """
#     if path_is_file(path=path):
#         p = verify_input_path_is_file(path=path)
#         # open file
#         with open(p, 'r', encoding=encoding) as geojson_file:
#             json_object = json.loads(geojson_file.read())
#         json_name = p.stem
#     elif path_is_http(path=path, headers=http_headers):
#         p = verify_input_path_is_http(path=path, headers=http_headers)
#         http_req = open_http_path(path=p, headers=http_headers)
#         json_object = json.loads(http_req.read())
#         json_name = http_req.info().get_filename().replace('.json', "").replace('.geojson', '')
#
#     return json_object, json_name


def json_object_to_feature_generator(
        json_object,
):
    """
    Transform geojson FeatureCollection / Feature or Geometry to geoformat Feature

    :param json_object: geojson object transform to python dict
    :return: geoformat feature
    """
    # if it's a Feature Collection
    if json_object["type"] == "FeatureCollection":
        for i_feat, json_feature in enumerate(json_object["features"]):
            for feature in json_object_to_feature_generator(json_feature):
                if feature:
                    yield feature

    # if it's a Feature
    elif json_object["type"] == "Feature":
        feature = {}

        if "properties" in json_object:
            if json_object["properties"]:
                feature["attributes"] = dict(json_object["properties"])

        if "geometry" in json_object:
            if json_object["geometry"]:
                feature["geometry"] = dict(json_object["geometry"])

        if feature:
            yield feature

    # if it's a geometry
    elif json_object["type"].upper() in GEOFORMAT_GEOMETRY_TYPE:
        feature = {"geometry": dict(json_object)}

        if feature:
            yield feature
    else:
        raise Exception("no geojson compatible data")


def from_geojson_get_features_list(
        geojson_in_dict,
        field_name_filter,
        geometry_type_filter,
        bbox_filter,
        serialize,
        bbox_extent,
        feature_limit,
        feature_offset
):
    """
    From geojson transform to dict this function return a features list with some filter option(s)

    :param geojson_in_dict:
    :param field_name_filter:
    :param geometry_type_filter:
    :param bbox_filter:
    :param serialize:
    :param bbox_extent:
    :param feature_limit:
    :param feature_offset:
    :return: list with feature(s) inside
    """
    # create feature generator from file
    features_generator = json_object_to_feature_generator(
        json_object=geojson_in_dict
    )
    # get features list
    features_list = [
        feature for feature in features_filter(
            geolayer_feature_list_or_generator=features_generator,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            serialize=serialize,
            bbox_extent=bbox_extent,
            feature_limit=feature_limit,
            feature_offset=feature_offset,
        )
    ]

    return features_list


def geojson_to_geolayer(
    path,
    geolayer_name=None,
    field_name_filter=None,
    geometry_type_filter=None,
    bbox_extent=True,
    bbox_filter=None,
    serialize=False,
    feature_limit=None,
    feature_offset=None,
    force_field_conversion=False,
    crs=None,
    http_headers=None,
    encoding="utf8"
):
    """
    Convert geojson file to geolayer

    :param path: path to geojson file
    :param field_name_filter: field_name that we want to keep in geolayer (can be a list).
    :param geometry_type_filter: keep only features with geometry type in this variable (can be a list).
    :param bbox_extent: add "bbox" key in each features and "extent" key in geometry metadata.
    :param bbox_filter: keep only feature that intersects bbox (can be a list of bbox).
    :param serialize: True if features in geolayer are serialized (can reduce performance) / False if not.
    :param feature_limit: constrains the number of rows returned in output geolayer.
    :param feature_offset: skip feature before beginning to given line number.
    :param force_field_conversion:  True if you want to force value in field (can change field type) / False if you want
           to deduce field type without forcing field type.
    :param crs: epsg code for coordinates reference system.
    :param http_headers:
    :param encoding: string encoding default utf8

    :return geolayer: geolayer
    """

    # init and prepare variable
    field_name_filter = value_to_iterable_value(field_name_filter, output_iterable_type=list)
    geometry_type_filter = value_to_iterable_value(geometry_type_filter, output_iterable_type=set)
    bbox_filter = value_to_iterable_value(bbox_filter, output_iterable_type=tuple)

    # load data and get name
    data_geojson_raw, name = load_data(path=path, encoding=encoding, http_headers=http_headers)
    with open(data_geojson_raw.name, 'r') as geojson_data:
        data_geojson_in_dict = json.loads(geojson_data.read())
        geolayer_name = geolayer_name or name or GEOLAYER_DEFAULT_NAME
        if data_geojson_in_dict:
            # filter features
            features_list = from_geojson_get_features_list(
                geojson_in_dict=data_geojson_in_dict,
                field_name_filter=field_name_filter,
                geometry_type_filter=geometry_type_filter,
                bbox_filter=bbox_filter,
                serialize=None,
                bbox_extent=False,
                feature_limit=feature_limit,
                feature_offset=feature_offset
            )

            # create geolayer from filter feature list
            geolayer = feature_list_to_geolayer(
                feature_list=features_list,
                geolayer_name=geolayer_name,
                field_name_filter=None,
                force_field_conversion=force_field_conversion,
                geometry_type_filter=None,
                bbox_filter=None,
                bbox_extent=bbox_extent,
                crs=crs,
                serialize=serialize,
            )

            # add coordinate reference system
            if "crs" in data_geojson_in_dict and crs is None:
                if 'geometry_ref' in geolayer['metadata']:
                    crs_raw_value = data_geojson_in_dict['crs']['properties']['name'].split(':')[-1]
                    try:
                        crs = int(crs_raw_value)
                    except ValueError:
                        crs = proj_correspondance.get(crs_raw_value)

            if crs and 'geometry_ref' in geolayer['metadata']:
                geolayer['metadata']['geometry_ref']['crs'] = crs

        return geolayer



import copy
from typing import Union, List, Any

from geoformat.geoprocessing.measure.mesure_area import triangle_area
from geoformat.geoprocessing.connectors.operations import (
    coordinates_to_point
)

from geoformat.geoprocessing.geoparameters.boundaries import ccw_or_cw_boundary

from geoformat.conf.error_messages import import_pyproj_error

try:
    from pyproj import Transformer
    import_pyproj_success = True
except ImportError:
    import_pyproj_success = False


def format_coordinates(
        coordinates_list_tuple,
        format_to_type=list,
        precision=None,
        delete_duplicate_following_coordinates=False,
        delete_vertex_list=None,
        delete_collinear_vertex=False,
        in_crs=None,
        out_crs=None,
        translate=None,
):
    """
    This function allow to :
        - reproject coordinates to in_crs to another out_crs
        - round coordinates
        - delete duplicate following coordinates in
        - change type of coordinate's storage

    :param coordinates_list_tuple: tuple or list containing coordinates
    :param format_to_type:  storage data type (list (by default) or tuple)
    :param precision: desired number of decimal for coordinates
    :param delete_duplicate_following_coordinates: option to delete similar coordinates that follow each other
    :param delete_vertex_list: delete vertex if exists in coordinates - list of vertex
    :param delete_collinear_vertex: delete collinear vertex in coordinates
    :param in_crs: input coordinates EPSG.
    :param out_crs: output coordinates EPSG wanted.
    :param out_crs: output coordinates EPSG wanted.
    :param translate: if you want to translate coordinates on x or y axis. example (-1, 1) it will move coordinates
        to -1 on x and +1 on y.
    :return: formatted coordinates
    """

    # format delete_vertex_list
    if delete_vertex_list is not None:
        if isinstance(delete_vertex_list[0], tuple):
            delete_vertex_list = format_coordinates(coordinates_list_tuple=delete_vertex_list, format_to_type=list)
    if import_pyproj_success is True:
        # TODO add delete duplicated part
        if coordinates_list_tuple:
            output_coordinates_list_tuple = copy.deepcopy(coordinates_list_tuple)
            point_list = False
            if isinstance(output_coordinates_list_tuple[0], (float, int)):
                point_list = True
                output_coordinates_list_tuple = [output_coordinates_list_tuple]

            if isinstance(output_coordinates_list_tuple[0][0], (int, float)):
                # reproject if necessary
                if in_crs is not None and out_crs is not None:
                    x_coords, y_coords = zip(*output_coordinates_list_tuple)
                    transformer = Transformer.from_crs(f"EPSG:{in_crs}", f"EPSG:{out_crs}", always_xy=True)
                    transform_coords = transformer.transform(x_coords, y_coords)
                    output_coordinates_list_tuple = zip(*transform_coords)

                if precision is not None:
                    # change precision
                    output_coordinates_list_tuple = [[round(xyz, precision) for xyz in coordinates] for coordinates in
                                                     output_coordinates_list_tuple]

                if translate is not None:
                    output_coordinates_list_tuple = [[xyz + translate[i] for i, xyz in enumerate(coordinates)] for coordinates in
                                                     output_coordinates_list_tuple]
                # FORMAT coordinates storage type
                # reformat coordinates storage type (list or tuple)
                if format_to_type:
                    output_coordinates_list_tuple = format_to_type(
                        [format_to_type(coordinates) for coordinates in output_coordinates_list_tuple])
            else:

                output_coordinates_list_tuple = [None] * len(coordinates_list_tuple)
                # here we first format type of storage of coordinates and precision
                for i_coord, coordinates in enumerate(coordinates_list_tuple):
                    coord_tuple = format_coordinates(
                        coordinates_list_tuple=coordinates,
                        format_to_type=format_to_type,
                        precision=precision,
                        delete_duplicate_following_coordinates=delete_duplicate_following_coordinates,
                        delete_vertex_list=delete_vertex_list,
                        delete_collinear_vertex=delete_collinear_vertex,
                        in_crs=in_crs,
                        out_crs=out_crs,
                        translate=translate,
                    )
                    output_coordinates_list_tuple[i_coord] = coord_tuple

                if format_to_type:
                    output_coordinates_list_tuple = format_to_type(output_coordinates_list_tuple)

            # reformat coordinates list to point
            if point_list is True:
                output_coordinates_list_tuple = output_coordinates_list_tuple[0]

            # DELETE COORDINATES OR DUPLICATE FOLLOWING COORDINATES
            # in step before if we reformat precision we can create duplicate coordinates then if
            # delete_duplicate_following_coordinates is True we have to make a second scan of coordinates
            # if we are at level of part of geometry (list that contains coordinates in it)
            if (
                    (
                            delete_duplicate_following_coordinates is True
                            or delete_collinear_vertex is True
                            or (
                                    delete_vertex_list is not None
                                    and isinstance(output_coordinates_list_tuple, (list, tuple))
                                    and len(output_coordinates_list_tuple) > 0
                                    and isinstance(output_coordinates_list_tuple[0], (list, tuple))
                                    and len(output_coordinates_list_tuple[0]) > 0
                                    and isinstance(output_coordinates_list_tuple[0][0], (int, float))
                            )
                    )
            ):
                if delete_collinear_vertex is True:
                    # in this case we delete duplicate following coordinates first to avoid mix of duplicate and
                    # collinear point at same time. So fist we delete duplicate then we found collinear and delete them
                    output_coordinates_list_tuple = format_coordinates(
                        coordinates_list_tuple=output_coordinates_list_tuple,
                        delete_duplicate_following_coordinates=True,
                    )
                # scan list of coordinates or list of coordinates (don't scan coordinate list)
                # if coordinates
                retype_to_tuple = False
                # transform tuple to list (because we need to edit coordinates)
                if isinstance(output_coordinates_list_tuple, tuple):
                    retype_to_tuple = True
                    output_coordinates_list_tuple = list(output_coordinates_list_tuple)

                # try to find duplicate coordinates index
                coordinates_to_delete_idx_list = []
                for idx_inside_coord, inside_coordinates in enumerate(output_coordinates_list_tuple):
                    delete_coordinate = False # init delete vertex variable
                    if idx_inside_coord == 0:
                        coordinates_to_delete_idx_list = []
                    else:
                        if delete_duplicate_following_coordinates is True:
                            # if coordinates are same that previous index of coordinates is save in
                            # duplicate_coordinates_idx_list (we delete this coordinates after the loop)
                            if previous_coordinates == inside_coordinates:
                                delete_coordinate = True
                        if delete_collinear_vertex is True and delete_coordinate is False:
                            if 0 < idx_inside_coord < len(output_coordinates_list_tuple) - 1:
                                next_coordinates = output_coordinates_list_tuple[idx_inside_coord+1]
                                if triangle_area(previous_coordinates, inside_coordinates, next_coordinates) == 0:
                                    delete_coordinate = True

                    # delete coordinates if wanted
                    if delete_vertex_list and delete_coordinate is False:
                        if inside_coordinates in delete_vertex_list:
                            delete_coordinate = True

                    if delete_coordinate is True:
                        coordinates_to_delete_idx_list.append(idx_inside_coord)

                    previous_coordinates = inside_coordinates

                # delete vertex
                if coordinates_to_delete_idx_list:
                    for idx in reversed(coordinates_to_delete_idx_list):
                        del output_coordinates_list_tuple[idx]

                # retype to tuple if output_coordinates_list_tuple is originally a tuple
                if retype_to_tuple is True:
                    output_coordinates_list_tuple = tuple(output_coordinates_list_tuple)
        else:
            if format_to_type == list:
                output_coordinates_list_tuple = []
            else:
                output_coordinates_list_tuple = ()

    else:
        raise Exception(import_pyproj_error)

    return output_coordinates_list_tuple


def coordinates_to_2d_coordinates(coordinates_list):
    """
    Convert a coordinates list with x dimension to 2d dimension list

    :param coordinates_list: list of coordinates with 2 or more dimension
    :return: coordinates list with only 2 dimensions.
    """

    if coordinates_list:
        if isinstance(coordinates_list[0], (int, float)):
            new_coordinates = [coordinates_list[0], coordinates_list[1]]
        elif isinstance(coordinates_list[0], (tuple, list)):
            new_coordinates = [None] * len(coordinates_list)
            for i_coord, under_coordinates in enumerate(coordinates_list):
                new_coordinates[i_coord] = coordinates_to_2d_coordinates(under_coordinates)
        else:
            raise Exception('error your geometry in input is not correct')
    else:
        new_coordinates = []

    return new_coordinates


def coordinates_to_centroid(coordinates_list_tuple, precision=None):
    """
    Return the centroid of given coordinates list or tuple

    :param coordinates_list_tuple: (list or tuple) coordinates list or tuple
    :param precision: (int) precision of coordinates (number of decimal places)
    :return: (tuple) centroid
    """

    for i_point, point in enumerate(coordinates_to_point(coordinates_list_tuple)):
        if i_point == 0:
            mean_point = list(point)
        else:
            mean_point[0] += point[0]
            mean_point[1] += point[1]

    nb_coordinates = i_point + 1
    centroid = [mean_point[0] / nb_coordinates, mean_point[1] / nb_coordinates]

    if precision:
        centroid = format_coordinates(coordinates_list_tuple=centroid, precision=precision)

    return centroid


def separate_coordinates_dimensions(
    coordinates: Union[List[Union[float, List[Any]]], tuple]
) -> List[List[List[Union[float, int]]]]:
    """
    Separates the coordinates dimensions of coordinates into distinct lists while preserving
    the original structure and depth of the input data.

    This function supports coordinates in point, line, polygon, and multipolygon formats,
    including those with an arbitrary number of dimensions (e.g., x, y, z).

    Parameters:
    ----------
    coordinates : Union[List[Union[float, List[Any]]], tuple]
        A list or tuple of coordinates. Each coordinate can be a number (for a point),
        a list of numbers (for a line or polygon), or a list of lists of numbers
        (for a polygon with holes or a multipolygon).

    Returns:
    -------
    List[List[List[Union[float, None]]]]
        A list of lists of lists where each top-level list represents one dimension
        of the coordinates (e.g., all x's, all y's, etc.), with the original nesting
        structure preserved.

    """
    # Converts coordinates to a list if not already, to standardize processing.
    if not isinstance(coordinates[0], (tuple, list)):
        coordinates = [coordinates]

    # Initializes the structure that will contain the final result.
    result_structure = [[None] for _ in range(len(coordinates))]
    separated_dimensions = None

    for index, coord_group in enumerate(coordinates):
        if isinstance(coord_group[0], (tuple, list)):
            # If the coordinate group is itself a list, process it recursively.
            nested_dimensions = separate_coordinates_dimensions(coord_group)
            if not separated_dimensions:
                separated_dimensions = [
                    [list(subgroup) for subgroup in result_structure]
                    for _ in range(len(nested_dimensions))
                ]
            for dim_index, nested_coord in enumerate(nested_dimensions):
                separated_dimensions[dim_index][index] = nested_coord
        else:
            # For a linear coordinate group (no sub-lists), separate the dimensions directly.
            if not separated_dimensions:
                separated_dimensions = [
                    [list(subgroup) for subgroup in result_structure]
                    for _ in range(len(coord_group))
                ]
            for dim_index, value in enumerate(coord_group):
                separated_dimensions[dim_index][index][0] = value

    return separated_dimensions

def force_rhr_polygon_coordinates(coordinates):
    output_ring = [None] * len(coordinates)
    # loop on each ring
    for i_ring, ring_coordinates in enumerate(coordinates):
        ring_direction = ccw_or_cw_boundary(ring_coordinates)
        # for exterior ring
        if i_ring == 0:
            # if counter clock wise we reverse ring's coordinates
            if ring_direction == 'CCW':
                ring_coordinates.reverse()
        # for interrior ring
        else:
            # if clock wise we reverse ring's coordinates
            if ring_direction == 'CW':
                ring_coordinates.reverse()
        # saving ring
        output_ring[i_ring] = ring_coordinates

    return output_ring


from geoformat.geoprocessing.connectors.operations import coordinates_to_bbox


def segment_list_to_linestring(segment_list, bbox=True, segment_as_part=False):
    """
    Create a linestring from a list of segment.

    :param segment_list: list that contains segment(s) in it
    :param bbox: True if we want bbox
    :param segment_as_part:
                        True : each segment are treated as a part.
                        False : all segments are considered to be part of the same line (duplicate coordinates are
                        removed).
    :return: linestring or multi linestring
    """

    if segment_as_part and len(segment_list) > 1:
        linestring = {'type': 'MultiLineString', 'coordinates': segment_list}
    else:
        linestring = {'type': 'LineString', 'coordinates': []}

        linestring_coordinates = [None] * len(segment_list) * 2
        duplicate_coordinates_list = []
        for i_seg, segment in enumerate(segment_list):
            idx = i_seg * 2
            linestring_coordinates[idx] = segment[0]
            linestring_coordinates[idx+1] = segment[1]
            # verify if there is duplicate coordinates
            if linestring_coordinates[idx-1] == linestring_coordinates[idx]:
                duplicate_coordinates_list.append(idx)

        if duplicate_coordinates_list:
            # format coordinates to delete duplicate coordinates
            for idx in reversed(duplicate_coordinates_list):
                del linestring_coordinates[idx]

        # add coordinates to linestring
        linestring['coordinates'] = linestring_coordinates

    if bbox is True:
        linestring['bbox'] = coordinates_to_bbox(linestring['coordinates'])

    return linestring
